/**
 ** MPC_controllerExecTaskCodels.cc
 **
 ** Codels called by execution task MPC_controllerExecTask
 **
 ** Author: Vivien Delsart
 ** Date: Wed Feb 9 2011
 **
 **/

#include <portLib.h>

#include "server/MPC_controllerHeader.h"

#include <MPC/common.hpp>
#include <MPC/system.hpp>
#include <MPC/trajGen.hpp>
#include <MPC/controller.hpp>
#include <MPC/trajFollow.hpp>
#include <MPC/trajSel.hpp>
using namespace mpc;

#include <iostream>
#include <cstring>
using namespace std;

#define DIST_TARGET_REACHED 0.25

#define DEBUG 0
#define DISPLAY_TIME false

// trajectory selection options
#define MANATRAJSEL_SPG_ACC 9
#define MANATRAJSEL_SPG_AACC 9

#define MANATRAJSEL_SPG_ACC2 3
#define MANATRAJSEL_SPG_AACC2 3
#define MANATRAJSEL_SPG_JERK2 3
#define MANATRAJSEL_SPG_AJERK2 3

#define TF_DFLT_TIME_HORIZON 2.0

#define TRACK_CTRAJ_SAMPLING_NB 10
#define ADAPTIVE_TRACK_CTRAJ_SAMPLING true

ROBOTIC_SYS_T sys_type;              // Robotic system type ( dala, dala_simu   ->  MPC_TF_MANA_SYS
                                     //                       otherwise         ->  MPC_TF_UNDEFINED_SYS )

POSTER_ID pomPosterId;               // POM poster ID (containing the localization of the robot)
POSTER_ID demPosterId;               // DEM poster ID (containing the local Data Elevation Map of the environment surrounding the robot)
POSTER_ID strajPosterId;             // MPC_planner state trajectory poster ID (containing the trajectory followed by the robot)
POSTER_ID ctrajPosterId;             // MPC_planner control trajectory poster ID (containing the trajectory followed by the robot)

System* sys;                         // the robotic system managed by MPC

SimpleController* controller;        // the controller managed by MPC
TrajFollow* trajFollow;              // the trajectory following process
TrajSel* trajSel;                    // the obstacle avoidance process (by trajectory selection)

STraj* straj;                        // the trajectory followed
DEM* localDEM;                       // the local DEM of the robot

double distTgtReached;               // dist up to which the robot is considered to have reached the goal

double timeHorizon;                  // time horizon of the trajectory following process

#if DISPLAY_TIME
Timer timer_monitor;                 // timer used to monitor the real execution time of the module
#endif


/*------------------------------------------------------------------------
 *  Controller process tools
 */

/*
 *  \brief update the localization of the Mana robot
 *  \param the resulting state
 *  \return 0 if the process succeed, -1 else
 */
int updateManaLoc(ManaState &sres)
{
  // read the current location (POM poster)
  POM_POS pos;
  if(MPC_controllerPOM_POSPosterRead (pomPosterId, &pos) == ERROR)
    {
      return -1;
    }

  // get back the current velocity
  double curV = SDI_F->ctrl_send.v;
  double curW = SDI_F->ctrl_send.w;
  
  // build the current state
  sres.x() = pos.mainToOrigin.euler.x;
  sres.y() = pos.mainToOrigin.euler.y;
  sres.th() = pos.mainToOrigin.euler.yaw;
  sres.v() = curV;
  sres.w() = curW;  

  return 0;
}

/*
 *  \brief build a control trajectory from the input poster
 */
void buildManaCTraj(ManaPCTraj*& ctraj, struct timeval& date)
{

  // read the control trajectory poster
  MANA_PCTRAJ ctrajPoster;
  if(MPC_controllerMANA_PCTRAJPosterRead (ctrajPosterId, &ctrajPoster) == ERROR)
    {
      ctraj = NULL;
      return;
    }

  if(ctrajPoster.vel.n == 0 || ctrajPoster.avel.n == 0)
    {
      ctraj = NULL;
      return;
    }

  // convert the poster into a Mana control trajectory
  std::vector<std::vector<double> > vecVel;
  std::vector<std::vector<double> > vecAVel;

  for(int i=0; i<ctrajPoster.vel.n; i++)
    {
      vector<double> curPoly;
      curPoly.push_back(ctrajPoster.vel.tabItv[i].itv.binf);
      curPoly.push_back(ctrajPoster.vel.tabItv[i].itv.bsup);
      curPoly.push_back(ctrajPoster.vel.tabItv[i].poly.a);
      curPoly.push_back(ctrajPoster.vel.tabItv[i].poly.b);
      curPoly.push_back(ctrajPoster.vel.tabItv[i].poly.c);
      curPoly.push_back(ctrajPoster.vel.tabItv[i].poly.d);

      vecVel.push_back(curPoly);
    }

  for(int i=0; i<ctrajPoster.avel.n; i++)
    {
      vector<double> curPoly;
      curPoly.push_back(ctrajPoster.avel.tabItv[i].itv.binf);
      curPoly.push_back(ctrajPoster.avel.tabItv[i].itv.bsup);
      curPoly.push_back(ctrajPoster.avel.tabItv[i].poly.a);
      curPoly.push_back(ctrajPoster.avel.tabItv[i].poly.b);
      curPoly.push_back(ctrajPoster.avel.tabItv[i].poly.c);
      curPoly.push_back(ctrajPoster.avel.tabItv[i].poly.d);

      vecAVel.push_back(curPoly);
    }

  ctraj = new ManaPCTraj(vecVel, vecAVel);

  date.tv_sec = ctrajPoster.t0.sec;
  date.tv_usec = ctrajPoster.t0.usec;

//   cout << "ctrajPoster.t0.sec : " << ctrajPoster.t0.sec << "  ctrajPoster.t0.usec : " << ctrajPoster.t0.usec << "   date.tv_sec : " << date.tv_sec << "  date.tv_usec : " << date.tv_usec << endl;

//   cout << "++++++++++++++++++++++++++++++++++++++++++++ CONTROLLER TRAJ : ++++++++++++++++++++++++++++++++++++++++++++" << endl;
//   ctraj->display();
//   cout << "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++" << endl;
  
  return;
}


/*------------------------------------------------------------------------
 *  Trajectory following process tools
 */

/*
 *  \brief build a Mana trajectory from the input poster
 *  \return the Mana state trajectory if the process succeed, NULL else
 */
ManaSTraj* buildManaSTraj()
{
  PL_MANA_STRAJ PL_straj;
  if(MPC_controllerPL_MANA_STRAJPosterRead (strajPosterId, &PL_straj) == ERROR)
    {
      return NULL;
    }

  ManaSTraj* newSTraj = new ManaSTraj;
  int N = PL_straj.nbPos;
  for(int i=0; i<N; i++)
    {
      ManaStateT curST(PL_straj.tab[i].x, 
		       PL_straj.tab[i].y, 
		       PL_straj.tab[i].th, 
		       PL_straj.tab[i].v, 
		       PL_straj.tab[i].w, 
		       PL_straj.tab[i].t);

      newSTraj->append(curST);
    }
  return newSTraj;
}

/*------------------------------------------------------------------------
 *
 * MPC_TF_ExecTaskInit  --  Initialization codel (fIDS, ...)
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */

STATUS
MPC_TF_ExecTaskInit(int *report)
{
//   // initialize the execution task components
//   ctrler_status = CTRLER_IS_STOPPED;

  sys_type = MPC_UNDEFINED_SYS;

  sys = NULL;
  controller = NULL;
  trajFollow = NULL;
  trajSel = NULL;
  straj = NULL;
  localDEM = NULL;

  distTgtReached = DIST_TARGET_REACHED;

  timeHorizon = 2.0;

  return OK;
}

/*------------------------------------------------------------------------
 *
 * MPC_TF_ExecTaskEnd  --  Termination codel
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */

STATUS
MPC_TF_ExecTaskEnd(void)
{
  return OK;
}

/*------------------------------------------------------------------------
 *
 * MPC_TF_ExecTaskPerm  --  1st codel of permanent activity (called BEFORE other activities)
 *
 * Description: 
 * 
 * Reports:      OK
 * 
 * Returns:    OK or ERROR
 */

STATUS
MPC_TF_ExecTaskPerm(int *report)
{
// #if DEBUG
//   if(controller != NULL)
//     {
//       // get back the state trajectory followed
//       const ManaPSTraj* pstraj = ((ManaSimpleController*)controller)->getPSTraj();
      
//       if(pstraj != NULL)
// 	{
// 	  // sample the state trajectory
// 	  ManaSTraj* straj = pstraj->sample(PL_PLAN_MAXSIZE);

// 	  if(straj != NULL)
// 	    {
// 	      // fill the path poster
// 	      for(int i=0; i<SDI_F->pathXY.nbPos; i++)
// 		{
// 		  ManaStateT curST = straj->at(i);

// 		  SDI_F->pathXY.tab[i].x = curST.x();
// 		  SDI_F->pathXY.tab[i].y = curST.y();
// 		}
// 	      SDI_F->pathXY.nbPos = straj->size();

// 	      delete straj;
// 	    }
// 	  delete pstraj;
// 	}
//     }
// #endif

  return OK;
}


/*------------------------------------------------------------------------
 * Init
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_controller_POSTER_NOT_FOUND
 */

/* MPC_InitMain  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_InitMain(INIT_I *Init_I, int *report)
{
  // check for the localization poster (POM)
  if (posterFind(Init_I->pomPoster, &pomPosterId) == ERROR) 
    {
      cerr << "MPC_controller: cannot find pom poster" << endl;
      *report = S_MPC_controller_POSTER_NOT_FOUND;
      return ETHER;
    }
#if DEBUG
  else cerr << "MPC_controller: pom poster found" << endl;
#endif

  // check for the Data Elevation Map poster (DEM)
  if (posterFind(Init_I->demPoster, &demPosterId) == ERROR) 
    {
      cerr << "MPC_controller: cannot find dem poster" << endl;
      *report = S_MPC_controller_POSTER_NOT_FOUND;
      return ETHER;
    }
#if DEBUG
  else cerr << "MPC_controller: dem poster found" << endl;
#endif

  // check for the state trajectory poster (MPC_planner ManaSTraj - Not mandatory)
  if (posterFind(Init_I->strajPoster, &strajPosterId) == ERROR) 
    {
      cerr << "MPC_controller: cannot find straj poster" << endl;
    }
#if DEBUG
  else cerr << "MPC_controller: straj poster found" << endl;
#endif

  /* Robotic system type */
  SDI_F->system_type = MPC_UNDEFINED_SYS;
    
  SDI_F->pathXY.nbPos = 0;

  /* Current goal location */
  SDI_F->goal_pos.x = 0;
  SDI_F->goal_pos.y = 0;
  SDI_F->goal_pos.theta = 0;

  SDI_F->ctrl_send.numRef = 0;
  SDI_F->ctrl_send.updatePeriod = 0;
  SDI_F->ctrl_send.v = 0.;
  SDI_F->ctrl_send.w = 0.;
  SDI_F->ctrl_send.vt = 0.;
  SDI_F->ctrl_send.vmax = 0.;
  SDI_F->ctrl_send.wmax = 0.;
  SDI_F->ctrl_send.linAccelMax = 1.; /* unused anyway */
  SDI_F->ctrl_send.angAccelMax = 1.;

  SDI_F->mode = CTRLER_NOT_RUNNING;

  SDI_F->ctraj.vel.n = 0;
  SDI_F->ctraj.avel.n = 0;
  SDI_F->ctraj.t0.sec = 0;
  SDI_F->ctraj.t0.usec = 0;
  SDI_F->ctraj.tf.sec = 0;
  SDI_F->ctraj.tf.usec = 0;

  SDI_F->ctraj_eval_time = 0.0;

//   // check for the robotic system type
//   if(strcmp(Init_I->robot, "dala") == 0 || 
//      strcmp(Init_I->robot, "dala-simu") == 0 )
//     {
// #if DEBUG
//       cerr << "MPC_controller_ExecTask: Dala system found" << endl;
// #endif
      
//     }
//   else
//     {
// #if DEBUG
//       cerr << "MPC_controller_ExecTask: Undefined system" << endl;
// #endif      
//       return ETHER;
//     }
    
  return ETHER;
}

/*------------------------------------------------------------------------
 * GoTo
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_controller_NO_PATH
 *              S_MPC_controller_BLOCKED
 *              S_MPC_controller_POSTER_READ
 */

/* MPC_GoToStart  -  codel START of GoTo
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_GoToStart(GOTO_I *Goto_I, int *report)
{
  // check if the robotic system has been defined
  if(sys_type == MPC_UNDEFINED_SYS)
    {
      cerr << "MPC_controller_ExecTask: Cannot start the GoTo process : Undefined robotic system." << endl;      
      *report = S_MPC_controller_SYS_UNDEFINED;
      return FAIL;
    }

  // check if the POM poster has been found
  if(pomPosterId == NULL)
    {
      cerr << "MPC_controller_ExecTask: Cannot start the GoTo process : POM poster has not been found." << endl;      
      *report = S_MPC_controller_POM_NOT_FOUND;
      return FAIL;
    }

  // Reset the controller with a simple controller for a Mana robotic system
  if(controller != NULL) delete controller;
#if DEBUG
//   controller = new ManaSimpleController((ManaSys*)sys, 1.0, 1);
  controller = new ManaSimpleController((ManaSys*)sys, 1.0);
#else
  controller = new ManaSimpleController((ManaSys*)sys, 1.0);
#endif

  cout << "GOTO PROCESS STARTED" << endl;

  // set the goal location
  SDI_F->goal_pos.x = Goto_I->goal.x;
  SDI_F->goal_pos.y = Goto_I->goal.y;
  SDI_F->goal_pos.theta = Goto_I->goal.theta;

  // set the current status as started
  SDI_F->mode = CTRLER_SIMPLE_CONTROLLER;

  *report = OK;
  return EXEC;

//   return MPC_TF_StartStart(Goto_I, report);
}

/* MPC_GoToExec  -  codel EXEC of GoTo
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_GoToExec(GOTO_I *Goto_I, int *report)
{
//   if(ctrler_status == CTRLER_IS_STARTED)
  if(SDI_F->mode == CTRLER_SIMPLE_CONTROLLER)
    {
      // update the Mana localization
      ManaState s1; 
      if(updateManaLoc(s1) != 0)
	{
	  // send a null control
	  SDI_F->ctrl_send.v = 0;
	  SDI_F->ctrl_send.w = 0;
	  return FAIL;
	}

      // get back the goal location and  build the goal state
      ManaState s2(SDI_F->goal_pos.x,
		   SDI_F->goal_pos.y,
		   0.0,
		   0.0,
		   0.0);

      // check if the goal position has been reached
      double dx = s2.x() - s1.x();
      double dy = s2.y() - s1.y();
      double dist = sqrt(dx*dx + dy*dy);
      if(dist < distTgtReached)
	{
	  // the goal position has been reached
// 	  ctrler_status = CTRLER_IS_STOPPED;  
	  SDI_F->mode = CTRLER_NOT_RUNNING;
	  SDI_F->ctrl_send.v = 0;
	  SDI_F->ctrl_send.w = 0;
	  SDI_F->ctrl_send.numRef++;

// #if DEBUG
	  cout << "THE GOAL HAS BEEN REACHED" << endl;
// #endif
	  return END;
	}

      // read the current DEM (??? poster)
      // ...

      // compute the control to apply
      ManaControl ctrl = ((ManaSimpleController*)controller)->computeControl(s1,s2);

      // write the control poster
      SDI_F->ctrl_send.v = ctrl.v();
      SDI_F->ctrl_send.w = ctrl.w();
      SDI_F->ctrl_send.numRef++;

#if DEBUG
      cout << "CONTROL : (" << SDI_F->ctrl_send.v << "," << SDI_F->ctrl_send.w << ")" << endl;
#endif
      *report = OK;
      return EXEC;
    }
  else
    {
      // send a null control
      SDI_F->ctrl_send.v = 0;
      SDI_F->ctrl_send.w = 0;
      SDI_F->ctrl_send.numRef++;

#if DEBUG
      cout << "CONTROL : (" << SDI_F->ctrl_send.v << "," << SDI_F->ctrl_send.w << ")" << endl;
#endif

      *report = OK;
      return ETHER;
    }
}

/* MPC_TF_GoToEnd  -  codel END of GoTo
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_TF_GoToEnd(GOTO_I *Goto_I, int *report)
{
  SDI_F->mode = CTRLER_NOT_RUNNING;

  // send a null control
  SDI_F->ctrl_send.v = 0;
  SDI_F->ctrl_send.w = 0;
  SDI_F->ctrl_send.numRef++;
  
#if DEBUG
  cout << "GOTO END CONTROL : (" << SDI_F->ctrl_send.v << "," << SDI_F->ctrl_send.w << ")" << endl;
#endif

  return ETHER;
}

/* MPC_TF_GoToFail  -  codel FAIL of GoTo
   Returns:  FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_TF_GoToFail(GOTO_I *Goto_I, int *report)
{
  SDI_F->mode = CTRLER_NOT_RUNNING;

  // send a null control
  SDI_F->ctrl_send.v = 0;
  SDI_F->ctrl_send.w = 0;
  SDI_F->ctrl_send.numRef++;
  
#if DEBUG
  cout << "GOTO FAIL CONTROL : (" << SDI_F->ctrl_send.v << "," << SDI_F->ctrl_send.w << ")" << endl;
#endif

  return ETHER;
}

/* MPC_GoToInter  -  codel INTER of GoTo
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_GoToInter(GOTO_I *Goto_I, int *report)
{
  SDI_F->mode = CTRLER_NOT_RUNNING;

  return ETHER;
}

/*------------------------------------------------------------------------
 * SetManaSys
 *
 * Description: 
 *
 * Reports:      OK
 */

/* MPC_TF_SetManaSysExec  -  codel EXEC of SetManaSys
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_TF_SetManaSysExec(SETMANASYS_I *SetManaSys_I, int *report)
{
  sys_type = MPC_MANA_SYS;

#if DEBUG
  cerr << "MPC_controller_ExecTask: Mana system configured" << endl;
#endif

  SDI_F->ctrl_send.vmax = SetManaSys_I->vmax;
  SDI_F->ctrl_send.wmax = SetManaSys_I->wmax;
  SDI_F->ctrl_send.linAccelMax = SetManaSys_I->accel;
  SDI_F->ctrl_send.angAccelMax = SetManaSys_I->wmax;

  // Reset the local system
  if(sys != NULL) delete sys;
  sys = new ManaSys(SetManaSys_I->vmax,
		    SetManaSys_I->accel,
		    SetManaSys_I->wmax,
		    SetManaSys_I->wmax,
		    SetManaSys_I->body_length,
		    SetManaSys_I->body_width,
		    false);

  return ETHER;
}


// /*------------------------------------------------------------------------
//  * FollowTraj
//  *
//  * Description: 
//  *
//  * Reports:      OK
//  *              S_MPC_controller_POM_NOT_FOUND
//  *              S_MPC_controller_DEM_NOT_FOUND
//  *              S_MPC_controller_SYS_UNDEFINED
//  */

// /* MPC_TF_Start  -  codel START of FollowTraj
//    Returns:  START EXEC END ETHER FAIL ZOMBIE */
// ACTIVITY_EVENT
// MPC_TF_Start(float *TimeHorizon_I, int *report)
// {
//   SDI_F->mode = CTRLER_TRAJECTORY_FOLLOWING;

//   // check if the robotic system has been defined
//   if(sys_type == MPC_UNDEFINED_SYS)
//     {
//       cerr << "MPC_controller: Cannot start the FollowTraj process : Undefined robotic system." << endl;      
//       *report = S_MPC_controller_SYS_UNDEFINED;
//       return FAIL;
//     }

//   // check if the POM poster has been found
//   if(pomPosterId == NULL)
//     {
//       cerr << "MPC_controller: Cannot start the FollowTraj process : POM poster has not been found." << endl;      
//       *report = S_MPC_controller_POM_NOT_FOUND;
//       return FAIL;
//     }

//   // check if the followed trajectory poster has been found (STRAJ)
//   if(strajPosterId == NULL)
//     {
//       cerr << "MPC_controller: Cannot start the FollowTraj process : STRAJ poster has not been found." << endl;      
//       *report = S_MPC_controller_STRAJ_NOT_FOUND;
//       return FAIL;
//     }

//   cout << "TRAJECTORY FOLLOWING CHECKING DONE" << endl;

//   // build a trajectory from the straj poster
//   if(straj != NULL) delete straj;
//   straj = buildManaSTraj();
//   if(straj == NULL)
//     {
//       cerr << "MPC_controller: Cannot start the FollowTraj process : The state trajectory building failed." << endl;      
//       return FAIL;
//     }
// #if DEBUG
//   else 
//     {
//       cout << "Followed state trajectory : " << endl;
//       straj->display(); 
//       cout << endl;
//     }
// #endif

// #if DEBUG
//   cout << "FOLLOWED TRAJECTORY RETRIEVED" << endl;
// #endif 

//   // create a trajectory following process to follow the new trajectory
//   if(trajFollow != NULL) delete trajFollow;
//   trajFollow = new ManaTrajFollow((ManaSys*)sys, (ManaSTraj*)straj);

//   // check if the input time horizon is consistent
//   if(*TimeHorizon_I > 0)
//     {
//       // set the time horizon of the trajectory following process
//       trajFollow->dtReach() = *TimeHorizon_I;
//     }
//   else trajFollow->dtReach() = TF_DFLT_TIME_HORIZON;
//   cout << "TF time horizon : " << trajFollow->dtReach() << endl;

//   trajFollow->start();

//   // set the current status as started
//   SDI_F->mode = CTRLER_TRAJECTORY_FOLLOWING;

// #if DISPLAY_TIME
//   timer_monitor.reinit();
//   timer_monitor.start();
// #endif

//   cout << "TRAJECTORY FOLLOWING PROCESS STARTED" << endl;

//   *report = OK;
//   return EXEC;
// }

// /* MPC_TF_Main  -  codel EXEC of FollowTraj
//    Returns:  EXEC END ETHER FAIL ZOMBIE */
// ACTIVITY_EVENT
// MPC_TF_Main(float *TimeHorizon_I, int *report)
// {
// #if DISPLAY_TIME
//   cout << "MPC_controller::FollowTraj -> Elapsed time : " << timer_monitor.elapsedTime() << "   period : " << timer_monitor.avgPeriod() << "   frequency : " << timer_monitor.avgFrequency() << endl;
// #endif

//   // update the Mana localization
//   ManaState curS; 
//   if(updateManaLoc(curS) != 0)
//     {
//       // send a null control
//       SDI_F->ctrl_send.v = 0;
//       SDI_F->ctrl_send.w = 0;
//       return FAIL;
//     }

// //   // get back the current time and build the corresponding state-time
// //   double time = timer.totalElapsedTime();
// //   ManaStateT curST(curS,time);

// //   cout << "CURRENT STATE-TIME : "; curST.display(); cout << endl;

//   // get back the control to apply to follow the trajectory
//   ManaControl ctrl;
//   int res = -1;
//   try
//     {
// //       res = ((ManaTrajFollow*)trajFollow)->getControl(curST, curST.t(), 0.5, ctrl);
//       res = ((ManaTrajFollow*)trajFollow)->getControl(curS, ctrl);
//     }
//   catch(std::string str)
//     {
//       cout << "str : " << str << endl;
//       exit(1);
//     }
  
//   if(res != 0)
//     {
//       SDI_F->ctrl_send.v = 0;
//       SDI_F->ctrl_send.w = 0;
//       SDI_F->ctrl_send.numRef++;
// #if DEBUG
//       cout << "WARNING : THE TRAJECTORY GENERATION PROCESS FAILED TO JOIN THE FOLLOWED TRAJECTORY!" << endl;
// #endif
//       return EXEC;
//     }

//   else if(trajFollow->endTraj())
//     {
//       // the goal position has been reached
//       SDI_F->mode = CTRLER_NOT_RUNNING;
//       SDI_F->ctrl_send.v = 0;
//       SDI_F->ctrl_send.w = 0;
//       SDI_F->ctrl_send.numRef++;
// // #if DEBUG
//       cout << "THE END OF THE TRAJECTORY HAS BEEN REACHED" << endl;
// // #endif
//       *report = OK;
//       return END;
//     }

//   else
//     {
//       // send the control to the robot
//       SDI_F->ctrl_send.v = ctrl.v();
//       SDI_F->ctrl_send.w = ctrl.w();
//       SDI_F->ctrl_send.numRef++;
// #if DEBUG
//       cout << "CONTROL : (" << SDI_F->ctrl_send.v << "," << SDI_F->ctrl_send.w << ")" << endl;
// #endif
//       *report = OK;
//       return EXEC;
//     }
// }

// /* MPC_TF_End  -  codel END of FollowTraj
//    Returns:  END ETHER FAIL ZOMBIE */
// ACTIVITY_EVENT
// MPC_TF_End(float *TimeHorizon_I, int *report)
// {
//   // set the current status as stopped
//   SDI_F->mode = CTRLER_NOT_RUNNING;

//   // stops the trajectory following process
//   trajFollow->stop();

// #if DISPLAY_TIME
//   timer_monitor.stop();
// #endif

//   return ETHER;
// }

// /* MPC_TF_Fail  -  codel FAIL of FollowTraj
//    Returns:  FAIL ZOMBIE */
// ACTIVITY_EVENT
// MPC_TF_Fail(float *TimeHorizon_I, int *report)
// {
//   // set the current status as stopped
//   SDI_F->mode = CTRLER_NOT_RUNNING;

//   // stops the trajectory following process
//   trajFollow->stop();

// #if DISPLAY_TIME
//   timer_monitor.stop();
// #endif

//   return ETHER;
// }


/*------------------------------------------------------------------------
 * Stop
 *
 * Description: 
 *
 * Reports:      OK
 */

/* MPC_StopExec  -  codel EXEC of Stop
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_StopExec(int *report)
{
  // set the current status as stopped
  SDI_F->mode = CTRLER_NOT_RUNNING;

  cout << "CONTROLLER PROCESS STOPPED" << endl;

  *report = OK;
  return ETHER;
}


/*------------------------------------------------------------------------
 * TrackCTraj
 *
 * Description: 
 *
 * Reports:      OK
 */

/* TrackCTrajStart  -  codel START of TrackCTraj
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
TrackCTrajStart(TRACKCTRAJ_I *TrackCTraj_I, int *report)
{
  if(SDI_F->mode != CTRLER_NOT_RUNNING)
    {
      cout << "MPC_controller StartCTrajTrack -> Cannot start the control trajectory tracking process. The controller is already running." << endl;
      *report = S_MPC_controller_ALREADY_RUNNING;
      return ETHER;
    }

  // check if the robotic system has been defined
  if(sys_type == MPC_UNDEFINED_SYS)
    {
      cerr << "MPC_controller StartCTrajTrack -> Cannot start the control trajectory tracking process : Undefined robotic system." << endl;      
      *report = S_MPC_controller_SYS_UNDEFINED;
      return ETHER;
    }

  // check if the POM poster has been found
  if(pomPosterId == NULL)
    {
      cerr << "MPC_controller StartCTrajTrack -> Cannot start the control trajectory tracking process : POM poster has not been found." << endl;      
      *report = S_MPC_controller_POM_NOT_FOUND;
      return ETHER;
    }

  // check if the MANA_PCTRAJ poster has been found
  if (posterFind(TrackCTraj_I->ctrajPoster, &ctrajPosterId) == ERROR) 
    {
      cerr << "MPC_controller StartCTrajTrack -> Cannot start the control trajectory tracking process : MANA_PCTRAJ poster has not been found." << endl;      
      *report = S_MPC_controller_PCTRAJ_NOT_FOUND;
      return ETHER;
    }

  SDI_F->ctraj_eval_time = TrackCTraj_I->ctraj_eval_time;

#if DISPLAY_TIME
  timer_monitor.reinit();
  timer_monitor.start();
#endif

  cerr << "TRACK CTRAJ PROCESS STARTED" << endl;

  SDI_F->mode = CTRLER_CTRAJ_TRACKING;

  return EXEC;
}

/* TrackCTrajMain  -  codel EXEC of TrackCTraj
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
TrackCTrajMain(TRACKCTRAJ_I *TrackCTraj_I, int *report)
{
#if DISPLAY_TIME
  cout << "MPC_controller::TrackCTraj -> Elapsed time : " << timer_monitor.elapsedTime() << "   period : " << timer_monitor.avgPeriod() << "   frequency : " << timer_monitor.avgFrequency() << endl;
#endif

  // Build a control trajectory from the poster
  ManaPCTraj* ctraj;
  struct timeval date;
  buildManaCTraj(ctraj, date);  

  if(ctraj == NULL)
    {
//       cerr << "MPC_controller TrackCTraj : The control trajectory is NULL." << endl;

      // set a null control
      SDI_F->ctrl_send.v = 0;
      SDI_F->ctrl_send.w = 0;
      SDI_F->ctrl_send.numRef++;      

#if DEBUG
      cerr << "CT CONTROL taj null : (" << SDI_F->ctrl_send.v << "," << SDI_F->ctrl_send.w << ")" << endl;
#endif

      return EXEC;
    }
//   else cerr << "MPC_controller TrackCTraj : The control trajectory is NOT NULL." << endl;

  // get back the current time
  struct timeval curTime;
  Timer::absoluteTime(curTime);

  double dt = Timer::computeElapsedTime(date, curTime);

//   cout << "date (ctraj) : " << date.tv_sec << " " << date.tv_usec << "   curTime : " <<  curTime.tv_sec << " " << curTime.tv_usec << "   dt : " << dt << endl;
//   cout << "dt : " << dt << "  ctraj->startTime() : " << ctraj->startTime() << "  ctraj->endTime() : " << ctraj->endTime() << "  SDI_F->ctraj_eval_time : " << SDI_F->ctraj_eval_time << endl;

  // check if the last computed trajectory is defined at the current time
//   if(curTime < ctraj->startTime() || curTime+0.5 > ctraj->endTime())
  if(dt < ctraj->startTime() || dt+SDI_F->ctraj_eval_time > ctraj->endTime())
    {
//       cerr << "MPC_controller TrackCTraj : The control trajectory is NOT DEFINED at the current time." << endl;

      // set a null control
      SDI_F->ctrl_send.v = 0;
      SDI_F->ctrl_send.w = 0;
      SDI_F->ctrl_send.numRef++;

#if DEBUG
      cerr << "CT CONTROL traj not defined : (" << SDI_F->ctrl_send.v << "," << SDI_F->ctrl_send.w << ")" << endl;
#endif
      return EXEC;
    }
//   else cerr << "MPC_controller TrackCTraj : The control trajectory is DEFINED at the current time." << endl;

  // compute the control to apply
//   ManaControl ctrl0 = ctraj->at(dt);
//   ManaControl ctrl1 = ctraj->at(dt + SDI_F->ctraj_eval_time);
//   ManaControl ctrl(0.5*(ctrl0.v() + ctrl1.v()), 
// 		   0.5*(ctrl0.w() + ctrl1.w()));
  ManaControl ctrl(0,0);
  if(ADAPTIVE_TRACK_CTRAJ_SAMPLING && TRACK_CTRAJ_SAMPLING_NB >= 1)
    {
      float dt2 = SDI_F->ctraj_eval_time / ((float)TRACK_CTRAJ_SAMPLING_NB);
      for(int i=0; i<=TRACK_CTRAJ_SAMPLING_NB; i++)
	{
	  ManaControl curCtrl = ctraj->at(dt + i*dt2);
	  ctrl += curCtrl;
	}
      ctrl /= ((double)TRACK_CTRAJ_SAMPLING_NB);

//       cout << "CT ADAPTIVE CONTROL : "; ctrl.display(); cout << "  dt : " << dt << endl;
    }
  else 
    {
      ctrl = ctraj->at(0.5);
//       cout << "CT STANDARD CONTROL : "; ctrl.display(); cout << endl;
    }

  // check for the system constraints
  if(((ManaSys*)sys)->canMoveBack())
    {
      if(ctrl.v() < -((ManaSys*)sys)->getVMax()) ctrl.v() = -((ManaSys*)sys)->getVMax();
    }
  else if(ctrl.v() < 0) ctrl.v() = 0;
  if(ctrl.v() > ((ManaSys*)sys)->getVMax()) ctrl.v() = ((ManaSys*)sys)->getVMax();

  if(ctrl.w() < -((ManaSys*)sys)->getAVMax()) ctrl.w() = -((ManaSys*)sys)->getAVMax();
  if(ctrl.w() > ((ManaSys*)sys)->getAVMax()) ctrl.w() = ((ManaSys*)sys)->getAVMax();
  
  // fill the control poster
  SDI_F->ctrl_send.v = ctrl.v();
  SDI_F->ctrl_send.w = ctrl.w();
  SDI_F->ctrl_send.numRef++;

#if DEBUG
  cerr << "CT CONTROL standard : (" << SDI_F->ctrl_send.v << "," << SDI_F->ctrl_send.w << ")" << endl;
#endif

  return EXEC;
}

/* TrackCTrajEnd  -  codel END of TrackCTraj
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
TrackCTrajEnd(TRACKCTRAJ_I *TrackCTraj_I, int *report)
{
  // set a null control
  SDI_F->ctrl_send.v = 0;
  SDI_F->ctrl_send.w = 0;
  SDI_F->ctrl_send.numRef++;

  // set the current status as stopped
  SDI_F->mode = CTRLER_NOT_RUNNING;

#if DISPLAY_TIME
  timer_monitor.stop();
#endif

  cerr << "TrackCTrajEnd : END" << endl;

  return ETHER;
}

/* TrackCTrajInter  -  codel INTER of TrackCTraj
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
TrackCTrajInter(TRACKCTRAJ_I *TrackCTraj_I, int *report)
{
  // set a null control
  SDI_F->ctrl_send.v = 0;
  SDI_F->ctrl_send.w = 0;
  SDI_F->ctrl_send.numRef++;

  // set the current status as stopped
  SDI_F->mode = CTRLER_NOT_RUNNING;

#if DISPLAY_TIME
  timer_monitor.stop();
#endif

  cerr << "TrackCTrajInter : INTER" << endl;

  return ETHER;
}
